"""Utilities for working with master"""
import json

from datetime import timedelta
from shakedown import *
from shakedown.cli.helpers import *
from shakedown.dcos.zookeeper import get_zk_node_children, get_zk_node_data
from shakedown.dcos.agent import kill_process_from_pid_file_on_host
from shakedown.dcos.spinner import time_wait
from shakedown.dcos import network

DISABLE_MASTER_INCOMING = "-I INPUT -p tcp --dport 5050 -j REJECT"
DISABLE_MASTER_OUTGOING = "-I OUTPUT -p tcp --sport 5050 -j REJECT"


def partition_master(incoming=True, outgoing=True):
    """ Partition master's port alone. To keep DC/OS cluster running.

    :param incoming: Partition incoming traffic to master process. Default True.
    :param outgoing: Partition outgoing traffic from master process. Default True.
    """

    echo('Partitioning master. Incoming:{} | Outgoing:{}'.format(incoming, outgoing))

    network.save_iptables(shakedown.master_ip())
    network.flush_all_rules(shakedown.master_ip())
    network.allow_all_traffic(shakedown.master_ip())

    if incoming and outgoing:
        network.run_iptables(shakedown.master_ip(), DISABLE_MASTER_INCOMING)
        network.run_iptables(shakedown.master_ip(), DISABLE_MASTER_OUTGOING)
    elif incoming:
        network.run_iptables(shakedown.master_ip(), DISABLE_MASTER_INCOMING)
    elif outgoing:
        network.run_iptables(shakedown.master_ip(), DISABLE_MASTER_OUTGOING)
    else:
        pass


def reconnect_master():
    """ Reconnect a previously partitioned master to the network
    """
    network.restore_iptables(shakedown.master_ip())


def restart_master_node():
    """ Restarts the master node
    """

    run_command_on_master("sudo /sbin/shutdown -r now")


def systemctl_master(command='restart'):
    """ Used to start, stop or restart the master process
    """
    run_command_on_master('sudo systemctl {} dcos-mesos-master'.format(command))


def mesos_available_predicate():
    url = master_url()
    try:
        response = http.get(url)
        return response.status_code == 200
    except Exception as e:
        return False


def wait_for_mesos_endpoint(timeout_sec=timedelta(minutes=5).total_seconds()):
    """Checks the service url if available it returns true, on expiration
    it returns false"""

    return time_wait(lambda: mesos_available_predicate(), timeout_seconds=timeout_sec)


def __mesos_zk_nodes():
    """ Returns all the children nodes under /mesos in zk
    """
    return get_zk_node_children('/mesos')


def __master_zk_nodes_keys():
    """ The masters can be registered in zk with arbitrary ids which start with
        `json.info_`.  This provides a list of all master keys.
    """
    master_zk = []
    for node in __mesos_zk_nodes():
        if 'json.info' in node['title']:
            master_zk.append(node['key'])

    return master_zk


def get_all_masters():
    """ Returns the json object that represents each of the masters.
    """
    masters = []
    for master in __master_zk_nodes_keys():
        master_zk_str = get_zk_node_data(master)['str']
        masters.append(json.loads(master_zk_str))

    return masters


def master_leader_ip():
    """Returns the private IP of the mesos master leader.
    In a multi-master cluster this may not map to the public IP of the master_ip.
    """
    return dcos_dns_lookup('leader.mesos')[0]['ip']


def get_all_master_ips():
    """ Returns a list of IPs for the masters
    """
    ips = []
    for master in get_all_masters():
        ips.append(master['hostname'])

    return ips


def required_masters(count):
    """ Returns True if the number of private agents is equal to or greater than
    the count.  This is useful in using pytest skipif such as:
    `pytest.mark.skipif('required_masters(3)')` which will skip the test if
    the number of masters is only 1.

    :param count: the number of required masters.
    """
    master_count = len(get_all_masters())
    # reverse logic (skip if less than count)
    # returns True if less than count
    return master_count < count


def masters(count=1):
    return pytest.mark.skipif('required_masters({})'.format(count))


def start_master_http_service(port=7777, pid_file='python_http.pid'):
    """ Starts a http service on the master leader.  The main purpose is to serve
    up artifacts for launched test applications.   This is commonly used in combination
    with copying tests or artifacts to the leader than configuring the messos task
    to fetch from http://master.mesos:7777/artifact.tar (becareful in a multi-master env)

    :param port: port to use for the http service
    :return: pid_file
    """
    shakedown.run_command_on_master(
        'nohup /opt/mesosphere/bin/python -m http.server {} > http.log 2>&1 & '
        'echo $! > {}'.format(port, pid_file))
    return pid_file


@contextlib.contextmanager
def master_http_service(port=7777):
    pid_file = start_master_http_service(port)
    yield
    kill_process_from_pid_file_on_host(shakedown.master_ip(), pid_file)


@contextlib.contextmanager
def disconnected_master(incoming=True, outgoing=True):

    partition_master(incoming, outgoing)
    try:
        yield
    finally:
        # return config to previous state
        reconnect_master()



"""Utilities for work with docker commands and images"""

import json
import os

from datetime import timedelta
import shakedown

from dcos import marathon


def create_docker_credentials_file(
        username,
        password,
        file_name='docker.tar.gz'):
    """ Create a docker credentials file.
        Docker username and password are used to create a `{file_name}`
        with `.docker/config.json` containing the credentials.

        :param username: docker username
        :type username: str
        :param password: docker password
        :type password: str
        :param file_name: credentials file name `docker.tar.gz` by default
        :type file_name: str
    """

    import base64
    auth_hash = base64.b64encode(
        '{}:{}'.format(username, password).encode()).decode()

    config_json = {
        "auths": {
            "https://index.docker.io/v1/": {"auth": auth_hash}
        }
    }

    config_json_filename = 'config.json'
    # Write config.json to file
    with open(config_json_filename, 'w') as f:
        json.dump(config_json, f, indent=4)

    try:
        # Create a docker.tar.gz
        import tarfile
        with tarfile.open(file_name, 'w:gz') as tar:
            tar.add(config_json_filename, arcname='.docker/config.json')
            tar.close()
    except Exception as e:
        print('Failed to create a docker credentils file {}'.format(e))
        raise e
    finally:
        os.remove(config_json_filename)


def __distribute_docker_credentials_file(file_name='docker.tar.gz'):
    """ Create and copy docker credentials file to passed `{agents}`.
        Used to access private docker repositories in tests.
    """
    # Upload docker.tar.gz to all private agents
    for host in shakedown.get_private_agents():
        shakedown.copy_file(host, file_name)


def distribute_docker_credentials_to_private_agents(
        username,
        password,
        file_name='docker.tar.gz'):
    """ Create and distributes a docker credentials file to all private agents

        :param username: docker username
        :type username: str
        :param password: docker password
        :type password: str
        :param file_name: credentials file name `docker.tar.gz` by default
        :type file_name: str
    """

    create_docker_credentials_file(username, password, file_name)

    try:
        __distribute_docker_credentials_file()
    finally:
        os.remove(file_name)


def prefetch_docker_image_on_private_agents(
        image,
        timeout=timedelta(minutes=5).total_seconds()):
    """ Given a docker image. An app with the image is scale across the private
        agents to ensure that the image is prefetched to all nodes.

        :param image: docker image name
        :type image: str
        :param timeout: timeout for deployment wait in secs (default: 5m)
        :type password: int
    """
    agents = len(shakedown.get_private_agents())
    app = {
        "id": "/prefetch",
        "instances": agents,
        "container": {
            "type": "DOCKER",
            "docker": {"image": image}
        },
        "cpus": 0.1,
        "mem": 128
    }

    client = marathon.create_client()
    client.add_app(app)

    shakedown.deployment_wait(timeout)

    shakedown.delete_all_apps()
    shakedown.deployment_wait(timeout)



from dcos import util
import time as time_module
import traceback

import shakedown

from inspect import currentframe, getargvalues, getsource, getouterframes

logger = util.get_logger(__name__)


def wait_for(
        predicate,
        timeout_seconds=120,
        sleep_seconds=1,
        ignore_exceptions=True,
        inverse_predicate=False,
        noisy=False,
        required_consecutive_success_count=1):
    """ waits or spins for a predicate, returning the result.
        Predicate is a function that returns a truthy or falsy value.
        An exception in the function will be returned.
        A timeout will throw a TimeoutExpired Exception.

    """
    count = 0
    start_time = time_module.time()
    timeout = Deadline.create_deadline(timeout_seconds)
    while True:
        try:
            result = predicate()
        except Exception as e:
            if ignore_exceptions:
                if noisy:
                    logger.exception("Ignoring error during wait.")
            else:
                count = 0
                raise  # preserve original stack
        else:
            if (not inverse_predicate and result) or (inverse_predicate and not result):
                count = count + 1
            if count >= required_consecutive_success_count:
                return result

        if timeout.is_expired():
            funname = __stringify_predicate(predicate)
            raise TimeoutExpired(timeout_seconds, funname)
        if noisy:
            header = '{}[{}/{}]'.format(
                shakedown.cli.helpers.fchr('>>'),
                pretty_duration(time_module.time() - start_time),
                pretty_duration(timeout_seconds)
            )
            if required_consecutive_success_count > 1:
                header = '{} [{} of {} times]'.format(
                    header,
                    count,
                    required_consecutive_success_count)
            print('{} spinning...'.format(header))
        time_module.sleep(sleep_seconds)


def __stringify_predicate(predicate):
    """ Reflection of function name and parameters of the predicate being used.
    """
    funname = getsource(predicate).strip().split(' ')[2].rstrip(',')
    params = 'None'

    # if args dig in the stack
    if '()' not in funname:
        stack = getouterframes(currentframe())
        for frame in range(0, len(stack)):
            if funname in str(stack[frame]):
                _, _, _, params = getargvalues(stack[frame][0])

    return "function: {} params: {}".format(funname, params)


def time_wait(
    predicate,
    timeout_seconds=120,
    sleep_seconds=1,
    ignore_exceptions=True,
    inverse_predicate=False,
    noisy=True,
    required_consecutive_success_count=1):
    """ waits or spins for a predicate and returns the time of the wait.
        An exception in the function will be returned.
        A timeout will throw a TimeoutExpired Exception.

    """
    start = time_module.time()
    wait_for(predicate, timeout_seconds, sleep_seconds, ignore_exceptions, inverse_predicate, noisy, required_consecutive_success_count)
    return elapse_time(start)


def wait_while_exceptions(
        predicate,
        timeout_seconds=120,
        sleep_seconds=1,
        noisy=False):
    """ waits for a predicate, ignoring exceptions, returning the result.
        Predicate is a function.
        Exceptions will trigger the sleep and retry; any non-exception result
        will be returned.
        A timeout will throw a TimeoutExpired Exception.
    """
    start_time = time_module.time()
    timeout = Deadline.create_deadline(timeout_seconds)
    while True:
        try:
            result = predicate()
            return result
        except Exception as e:
            if noisy:
                logger.exception("Ignoring error during wait.")

        if timeout.is_expired():
            funname = __stringify_predicate(predicate)
            raise TimeoutExpired(timeout_seconds, funname)
        if noisy:
            header = '{}[{}/{}]'.format(
                shakedown.cli.helpers.fchr('>>'),
                pretty_duration(time_module.time() - start_time),
                pretty_duration(timeout_seconds)
            )
            print('{} spinning...'.format(header))
        time_module.sleep(sleep_seconds)


def elapse_time(start, end=None, precision=3):
    """ Simple time calculation utility.   Given a start time, it will provide an elapse time.
    """
    if end is None:
        end = time_module.time()
    return round(end-start, precision)


def pretty_duration(seconds):
    """ Returns a user-friendly representation of the provided duration in seconds.
    For example: 62.8 => "1m2.8s", or 129837.8 => "2d12h4m57.8s"
    """
    if seconds is None:
        return ''
    ret = ''
    if seconds >= 86400:
        ret += '{:.0f}d'.format(int(seconds / 86400))
        seconds = seconds % 86400
    if seconds >= 3600:
        ret += '{:.0f}h'.format(int(seconds / 3600))
        seconds = seconds % 3600
    if seconds >= 60:
        ret += '{:.0f}m'.format(int(seconds / 60))
        seconds = seconds % 60
    if seconds > 0:
        ret += '{:.1f}s'.format(seconds)
    return ret


class Deadline(object):

    def is_expired(self):
        raise NotImplementedError()

    @staticmethod
    def create_deadline(seconds):
        if seconds is None:
            return Forever()
        return Within(seconds)


class Within(Deadline):

    def __init__(self, seconds):
        super(Within, self).__init__()
        self._deadline = time_module.time() + seconds

    def is_expired(self):
        return time_module.time() >= self._deadline


class Forever(Deadline):

    def is_expired(self):
        return False


class TimeoutExpired(Exception):
    def __init__(self, timeout_seconds, what):
        super(TimeoutExpired, self).__init__(timeout_seconds, what)
        self._timeout_seconds = timeout_seconds
        self._what = what

    def __str__(self):
        return "Timeout of {0} expired waiting for {1}".format(pretty_duration(self._timeout_seconds), self._what)

    def __repr__(self):
        return "{0}: {1}".format(type(self).__name__, self)

    def __unicode__(self):
        return u"Timeout of {0} expired waiting for {1}".format(pretty_duration(self._timeout_seconds), self._what)
